__title__ = 'mindsdb_sql'
__package_name__ = 'mindsdb_sql'
__version__ = '0.0.14'
__description__ = "Pure python SQL parser"
__email__ = "jorge@mindsdb.com"
__author__ = 'MindsDB Inc'
__github__ = 'https://github.com/mindsdb/mindsdb_sql'
__pypi__ = 'https://pypi.org/project/mindsdb_sql'
__license__ = 'GPL-3.0'
__copyright__ = 'Copyright 2021- mindsdb'
